from .datasets import CpcDataset, CpcIterableDataset
from .datasets import CpcV2Dataset, CpcV2IterableDataset

from .modules.cpc_module import CpcModule
from .modules.cpc_v2_module import CpcV2Module
